from dataclasses import dataclass, field
from typing import Optional, Any, Iterable, Callable, Union, TYPE_CHECKING
from types import CodeType
from collections.abc import Iterable
from functools import total_ordering
import re
from enum import Enum, auto

import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from matplotlib.lines import Line2D
from matplotlib.axes import Axes
from matplotlib.ticker import MultipleLocator
from matplotlib.transforms import ScaledTranslation
import numpy as np
from numpy.typing import ArrayLike

from .types import cm, Size

if TYPE_CHECKING:
    from .saxes import SAxes



try:
    from icecream import ic
except ImportError:  # Graceful fallback if IceCream isn't installed.
    ic = lambda *a: None if not a else (a[0] if len(a) == 1 else a)  # noqa

@dataclass
class SFigureModel:
    width: Size = Size(21.0 * cm)
    height: Size = Size(29.7 * cm)
    grid: Size = Size(0.5 * cm)
    grid_options: dict[str, Any] = field(default_factory=lambda: dict(alpha=0.2, lw=0.5))
    output_file: Optional[str] = None
    dpi: int = 300
    #action: FigAction = FigAction.INTERACT
        
class SFigure(SFigureModel):
    saxes: list["SAxes"]
    figure: Figure
    def __init__(self, *args, f: Optional[Figure] = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.saxes = []
        self.set_figure(f)
    def set_figure(self, f: Optional[Figure] = None):
        if f is None:
            f = plt.figure()
        self.figure = f
        self.figure.set_size_inches((self.width, self.height))
        self.draw_grid(**self.grid_options)
    
    def draw_grid(self, **kwargs):
        xx = np.arange(0, self.width+self.grid, self.grid)
        yy = np.arange(0, self.height+self.grid, self.grid)
        x_min = min(xx)
        x_max = max(xx)
        y_min = min(yy)
        y_max = max(yy)
        for x in xx:
            self.figure.add_artist(Line2D((x,x), (y_min, y_max), transform = self.figure.dpi_scale_trans, **kwargs))
        for y in yy:
            self.figure.add_artist(Line2D((x_min, x_max), (y,y), transform = self.figure.dpi_scale_trans, **kwargs))
        
    def add_axes(self, axes: "SAxes"):
        #axes.setup(self.figure, left, bottom)
        self.saxes.append(axes)
    
    def __enter__(self) -> "SFigure":
        return self
    def __exit__(self,exc_type, exc_val, exc_tb):
        #print("XXX in __exit__")
        #print(self.saxes)
        for sax in self.saxes:
            sax.finalize()
        if self.output_file is None:
            plt.show()
        else:
            self.figure.savefig(self.output_file, dpi=self.dpi)



from dataclasses import dataclass, field
from typing import Optional, Any, Callable, Union, ClassVar
from types import CodeType
from enum import Enum, unique

import numpy as np
from .converter import register_model, register_impl
try:
    from icecream import ic
except ImportError:  # Graceful fallback if IceCream isn't installed.
    ic = lambda *a: None if not a else (a[0] if len(a) == 1 else a)  # noqa

from .saxes import SAxes
from .utils.math_expression import MathExpression

@unique
class DiscontinuityBelongsTo(Enum):
    GREATER = "x>x0"
    SMALLER = "x<x0"
    NONE    = "none"
@dataclass 
class Discontinuity:
    x0: float
    belongs_to: DiscontinuityBelongsTo = DiscontinuityBelongsTo.GREATER

@register_model  
@dataclass
class SGraphModel:
    class_id: ClassVar[str] = "Graph"
    function: Union[MathExpression, list[MathExpression]]
    label: Optional[str] = None
    plot_args: dict[str, Any] = field(default_factory=dict)
    discontinuities: list[Discontinuity] = field(default_factory=list)
    condition   : Optional[MathExpression] = None
    max_y: float = 100
    min_y: float = -100
    var_prefix:str = ""
    def __post_init__(self):
        if not isinstance(self.function, list):
            self.function = [self.function]
        # print(self.function, type(self.function[0]))
        if self.label is None:
            self.label = None
            #self.label = self.function[0]

@register_impl
class SGraph(SGraphModel):
    _saxes: SAxes
    _cond_array: Optional[np.ndarray]
    _ys: list[np.ndarray]
    
    def __init__(self, saxes: SAxes, *args, **kwargs):
        self._saxes = saxes
        self._cond_array = None
        self._ys = [] 
        super().__init__(*args, **kwargs)
        self._current_color = "black"
        self.set_saxes(saxes)
        
            
        
    def set_saxes(self, saxes: SAxes):
        self._saxes = saxes
        self.x = self._saxes._x
        self._eval_functions()
        self._eval_condition()
        self.plot()
        
    def plot(self, **kwargs):
        args = self.plot_args.copy()
        args.update(kwargs)
        y = self._ys[0]
        if self._cond_array is not None:
            y[~self._cond_array] = np.nan
        
        line = self._saxes.plot(self.x, y, label=self.label, zorder=10, **args)[-1]
        self._current_color = line.get_color()
        l, = self.x.shape
        for dc in self.discontinuities:
            idx = int(np.argmin(np.abs(self.x - dc.x0)))
            if dc.belongs_to == DiscontinuityBelongsTo.NONE:
                idx1 = min(l-1, idx + 1)
                idx2 = max(0, idx - 1)
                self._saxes.plot([self.x[idx1]], [y[idx1]], "o", zorder=20, mfc="white", color=self._current_color)
                self._saxes.plot([self.x[idx2]], [y[idx2]], "o", zorder=20, mfc="white", color=self._current_color)
                continue
            elif dc.belongs_to == DiscontinuityBelongsTo.GREATER:
                idx2 = min(l-1, idx + 1)
                idx1 = max(0, idx - 1)
            else:
                idx1 = min(l-1, idx + 1)
                idx2 = max(0, idx - 1)
            self._saxes.plot([self.x[idx1]], [y[idx1]], "o", zorder=20, mfc="white", color=self._current_color)
            self._saxes.plot([self.x[idx2]], [y[idx2]], "o", zorder=20, color=self._current_color)
            
       
        
    def get_locals(self) -> dict[str, np.ndarray]:
        d1: dict[str, Any] =  dict(x=self.x, y=self._ys[0])
        d1.update(self._saxes.axes_variables)
        return d1
    
    def _eval_condition(self):
        if self.condition is None:
            return
        cond_array = self.condition.evaluate(self.get_locals())
        if not isinstance(cond_array, np.ndarray):
            print(f"Warning: condition '{self.condition}' does not evaluate to array.")
            return
        if not cond_array.dtype == np.bool_:
            print(f"Warning: condition '{self.condition}' does not evaluate to boolean expression")
            return
        self._cond_array = cond_array
    
    def _eval_functions(self):
        for i, f in enumerate(self.function):
            y = f.evaluate(dict(x=self.x), 
                           self._saxes.axes_functions)
            y = np.array(y, dtype=np.float_)
            y = np.broadcast_to(y, self.x.shape).copy()
            y[y>self.max_y] = np.inf
            y[y<self.min_y] = -np.inf
            if self._cond_array is not None:
                y[~self._cond_array] = np.nan
            for dc in self.discontinuities:
                idx = np.argmin(np.abs(self.x - dc.x0))
                y[idx] = np.inf
            self._ys.append(y)
            


@register_model  
class FillBetweenModel(SGraphModel):
    class_id: ClassVar[str] = "Area"
@register_impl
class FillBetween(SGraph, FillBetweenModel):
    def set_saxes(self, saxes: SAxes):
        self._saxes = saxes
        self.x = self._saxes._x
        self._eval_functions()
        self._eval_condition()
        y = self._ys[0].copy()
        if self._cond_array is not None:
            y[~self._cond_array] = np.nan
        y2 = self._ys[1].copy()
        if self._cond_array is not None:
            y2[~self._cond_array] = np.nan
        
        self._saxes.axes.fill_between(self.x, y, y2, 
                                      label=self.label, **self.plot_args)
        tmp_cond = self.condition
        self.condition=None
        y = self._ys[0]
        y2 = self._ys[1]
        idx = np.argwhere(np.diff(np.sign(y - y2))).flatten()
        self.intersects = self.x[idx]
        update_dict= {f"{self.var_prefix}sect_x_{i}": value for (i,value) in enumerate(self.intersects)}
        self._saxes.axes_variables.update(update_dict)
        update_dict= {f"{self.var_prefix}sect_y_{i}": value.evaluate(dict(x=self.intersects)) for (i,value) in enumerate(self.function)}
        self.condition = tmp_cond
        self._saxes.axes_variables.update(update_dict)
        
        
    def get_closest_intersect(self, x:float):
        #ic(self.intersects)
        return self.intersects[np.argmin(np.abs(self.intersects - x))]

    def get_locals(self) -> dict[str, np.ndarray]:
        d1: dict[str, Any] =  dict(x=self.x, y1=self._ys[0], y2=self._ys[1])
        d1.update(self._saxes.axes_variables)
        return d1




from dataclasses import dataclass, field
from typing import Optional, Any, ClassVar
import numpy as np

from .converter import register_model, register_impl
from .utils.math_expression import MathExpression
try:
    from icecream import ic
except ImportError:  # Graceful fallback if IceCream isn't installed.
    ic = lambda *a: None if not a else (a[0] if len(a) == 1 else a)  # noqa

from .saxes import SAxes
    
@register_model  
@dataclass
class TextModel:
    class_id: ClassVar[str] = "Text"   
    text: str
    x: MathExpression
    y: MathExpression
    rotation: MathExpression = MathExpression(0)
    text_args: dict[str, Any] = field(default_factory=lambda: dict(zorder=100))
@register_impl
class Text(TextModel):
    
    _saxes: SAxes
    def __init__(self, saxes: SAxes, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.set_saxes(saxes)
        # print(f"in {self.class_id}.__init__", self)
    def set_saxes(self, saxes: SAxes):
        self._saxes = saxes
        x = self._saxes.evaluate(self.x)
        y = self._saxes.evaluate(self.y)
        rotation = self._saxes.evaluate(self.rotation)
        self._saxes.axes.text(x,y, self.text, rotation=rotation, 
                                 **self.text_args)



from dataclasses import dataclass, field
from typing import Optional, Any, Callable, Union, ClassVar, TYPE_CHECKING
from types import CodeType
from enum import Enum, unique

import numpy as np
from .converter import register_model, register_impl
try:
    from icecream import ic
except ImportError:  # Graceful fallback if IceCream isn't installed.
    ic = lambda *a: None if not a else (a[0] if len(a) == 1 else a)  # noqa

if TYPE_CHECKING:
    from .saxes import SAxes
from .utils.math_expression import MathExpression

@register_model
@dataclass
class SFunctionModel:
    class_id: ClassVar[str] = "Function"
    function: MathExpression
    name: str
    
@register_impl
class SFunction(SFunctionModel):
    def __init__(self, saxes: "SAxes", *args, **kwargs):
        self._saxes = saxes
        super().__init__(*args, **kwargs)
        f_x = lambda x: self.function.evaluate(dict(x=x), force_reeval=True)
        self._saxes.axes_functions.update({self.name: f_x})



#%%
from .math_parser import parser, vardict_t

from dataclasses import dataclass, field
from typing import Any

class MathExpression(str):
    def __new__(cls, s):
        return super().__new__(cls, str(s))
    def __init__(self, s, **kwargs):
        super().__init__()
        self._evaluated = None
        self.value = s
        
    def evaluate(self, vars: vardict_t = None, 
                 functions: vardict_t = None, 
                 force_reeval=False):
        if vars is None:
            vars = {}
        if self._evaluated is None or force_reeval:
            self._evaluated = parser.evaluate_expression(self, vars, functions)
        return self._evaluated

# %%

# @dataclass
# class A:
#     m: MathExpression
#     b: int = 12
#     
#     def __post_init__(self):
#         self.m = MathExpression(self.m)
#     
# from cattrs.preconf.pyyaml import make_converter
# converter = make_converter()
# a = A(m="1+3*(3-2)")
# a1 = converter.structure(converter.unstructure(a), A)
# a1 == a

# %%



## Code in this file is heavily based on the fourFn example from pyparsing
## (https://github.com/pyparsing/pyparsing/blob/master/examples/fourFn.py)
## The original copiright notice is repoduced below.

## Changes: 
## - add an object-oriented interface
## - add boolean operators, trying to reproduce the python semantics of these operators
## - same for comparison operators
## - include (some) numpy ufuncs

# fourFn.py
#
# Demonstration of the pyparsing module, implementing a simple 4-function expression parser,
# with support for scientific notation, and symbols for e and pi.
# Extended to add exponentiation and simple built-in functions.
# Extended test cases, simplified pushFirst method.
# Removed unnecessary expr.suppress() call (thanks Nathaniel Peterson!), and added Group
# Changed fnumber to use a Regex, which is now the preferred method
# Reformatted to latest pypyparsing features, support multiple and variable args to functions
#
# Copyright 2003-2019 by Paul McGuire
#

# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
# 
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#%%
from pyparsing import (
    Literal,
    Word,
    Group,
    Forward,
    alphas,
    alphanums,
    Regex,
    ParseException,
    CaselessKeyword,
    Suppress,
    delimitedList,
)
import math
import operator
from typing import Any, Optional
import numpy as np

vardict_t = Optional[dict[str, Any]]

# map operator symbols to corresponding arithmetic operations
epsilon = 1e-12
opn = {
    "**": operator.pow,
    "+": operator.add,
    "-": operator.sub,
    "*": operator.mul,
    "/": operator.truediv,
    "^": operator.pow,
    "and": operator.and_,
    "or": operator.or_,
    "xor": operator.xor,
    "not": operator.not_,
    "~": operator.not_,
    "<=": operator.le,
    "<": operator.lt,
    ">=": operator.ge,
    ">": operator.gt,
    "==": operator.eq,
    "!=": operator.ne,
}

# NumPy ufuncs with one argument
fn = {key: val for (key, val) in np.__dict__.items() 
      if isinstance(val, np.ufunc) 
      if ("e->e" in val.types) or ("ee->e" in val.types)}
    
    
    
class MathParser:
    def __init__(self):
        self.exprStack = []
        self.bnf = None
        self.init_BNF()

    def push_first(self, toks):
        self.exprStack.append(toks[0])


    def push_unary_minus(self, toks):
        for t in toks:
            if t == "-":
                self.exprStack.append("unary -")
            else:
                break
    def push_unary_not(self, toks):
        for t in toks:
            if t in ["~", "not"]:
                self.exprStack.append("unary not")
            else:
                break
            

    def init_BNF(self):
        """
        expop   :: '^'
        multop  :: '*' | '/'
        addop   :: '+' | '-'
        integer :: ['+' | '-'] '0'..'9'+
        atom    :: PI | E | real | fn '(' expr ')' | '(' expr ')'
        factor  :: atom [ expop factor ]*
        term    :: factor [ multop factor ]*
        expr    :: term [ addop term ]*
        
        
        ... not covered in this grammar, but present in the code below:
        - boolean operators
        - comparison operators
        """
        if self.bnf is None:
            # use CaselessKeyword for e and pi, to avoid accidentally matching
            # functions that start with 'e' or 'pi' (such as 'exp'); Keyword
            # and CaselessKeyword only match whole words
            e = CaselessKeyword("E")
            pi = CaselessKeyword("PI")
            # fnumber = Combine(Word("+-"+nums, nums) +
            #                    Optional("." + Optional(Word(nums))) +
            #                    Optional(e + Word("+-"+nums, nums)))
            # or use provided pyparsing_common.number, but convert back to str:
            # fnumber = ppc.number().addParseAction(lambda t: str(t[0]))
            fnumber = Regex(r"[+-]?\d+(?:\.\d*)?(?:[eE][+-]?\d+)?")
            ident = Word(alphas, alphanums + "_$")

            plus, minus, mult, div = map(Literal, "+-*/")
            le, less, ge, greater, equal, nequal = map(Literal, ["<=","<",">=",">","=","!="])
            and_, or_, xor, not_, not_alt = map(Literal, ["and", "or", "xor", "not", "~"])
            lpar, rpar = map(Suppress, "()")
            logicop = and_ | or_ | xor 
            compop = le | less | ge | greater | equal | nequal 
            addop = plus | minus
            notop = not_ | not_alt
            multop = mult | div
            expop = Literal("^") | Literal("**")

            logicexpr = Forward()
            compexpr = Forward()
            expr_list = delimitedList(Group(logicexpr)) 
            # add parse action that replaces the function identifier with a (name, number of args) tuple
            def insert_fn_argcount_tuple(t):
                fn = t.pop(0)
                num_args = len(t[0])
                t.insert(0, (fn, num_args))

            fn_call = (ident + lpar - Group(expr_list) + rpar).setParseAction(
                insert_fn_argcount_tuple
            )
            atom = (
                addop[...] 
                + (
                    (fn_call | pi | e | fnumber | ident).setParseAction(self.push_first)
                    | Group(lpar + logicexpr + rpar)
                )
            ).setParseAction(self.push_unary_minus)

            # by defining exponentiation as "atom [ ^ factor ]..." instead of "atom [ ^ atom ]...", we get right-to-left
            # exponents, instead of left-to-right that is, 2^3^2 = 2^(3^2), not (2^3)^2.
            factor = Forward()
            factor <<= atom + (expop + factor).setParseAction(self.push_first)[...]
            term = factor + (multop + factor).setParseAction(self.push_first)[...]
            expr = term + (addop + term).setParseAction(self.push_first)[...]
            compexpr  = (
                notop[...] + 
                expr + (compop + expr).setParseAction(self.push_first)[...]
            ).setParseAction(self.push_unary_not)
            logicexpr <<= compexpr + (logicop + compexpr).setParseAction(self.push_first)[...]
            self.bnf = logicexpr




    @staticmethod
    def evaluate_stack(s, var_dict: vardict_t = None,
                       func_dict: vardict_t = None):
        if var_dict is None:
            var_dict = dict()
        if func_dict is None:
            func_dict = fn
        else: 
            func_dict = func_dict.copy()
            func_dict.update(fn)
        op, num_args = s.pop(), 0
        if isinstance(op, tuple):
            op, num_args = op
        if op == "unary -":
            return -MathParser.evaluate_stack(s, var_dict)
        if op == "unary not":
            return ~ MathParser.evaluate_stack(s, var_dict)
        if op in opn:
            # note: operands are pushed onto the stack in reverse order
            op2 = MathParser.evaluate_stack(s, var_dict)
            op1 = MathParser.evaluate_stack(s, var_dict)
            return opn[op](op1, op2)
        elif op.lower() == "pi":
            return math.pi  # 3.1415926535
        elif op.lower() == "e":
            return math.e  # 2.718281828
        elif op in func_dict:
            # note: args are pushed onto the stack in reverse order
            args = reversed([MathParser.evaluate_stack(s, var_dict) for _ in range(num_args)])
            return func_dict[op](*args)
        elif op.isidentifier():
            try:
                return var_dict[op]
            except KeyError:
                raise KeyError(f"unknown variable {op}")
        elif op[0].isalpha():
            raise Exception("invalid identifier '%s'" % op)
        else:
            # try to evaluate as int first, then as float if int fails
            try:
                return int(op)
            except ValueError:
                return float(op)
    def compile_expression(self, s: str):
        self.exprStack[:] = []
        if self.bnf is None:
            raise ValueError("This should not happen...")
        self.results = self.bnf.parseString(s, parseAll=True)
        return self.exprStack[:]
    def evaluate_compiled_expression(self, 
                                     expr_stack, 
                                     var_dict:vardict_t = None,
                                     func_dict:vardict_t = None):
        return self.evaluate_stack(expr_stack[:], var_dict, func_dict)
    def evaluate_expression(self, s: str, 
                            var_dict:vardict_t = None,
                             func_dict:vardict_t = None):
        stack = self.compile_expression(s)
        return self.evaluate_compiled_expression(stack, var_dict, func_dict)
    def evaluate_expression_(self, s: str, 
                             var_dict:vardict_t = None):
        if self.bnf is None:
            raise ValueError("This should not happen...")
        self.exprStack[:] = []
        self.results = self.bnf.parseString(s, parseAll=True)
        val = self.evaluate_stack(self.exprStack[:], var_dict)
        return val
    
    
parser=MathParser()

if __name__ == "__main__":
    import numpy as np
    p = MathParser()
    x = np.linspace(-1,1,10)
    print(p.evaluate_expression("sin(x_1222see)", dict(x_1222see=x)))


    def test(s, expected):
        try:
            val = p.evaluate_expression(s)
        except ParseException as pe:
            print(s, "failed parse:", str(pe))
        except Exception as e:
            print(s, "failed eval:", str(e), p.exprStack)
        else:
            if val == expected:
                print(s, "=", val, p.results, "=>", p.exprStack)
            else:
                1/0
                print(s + "!!!", val, "!=", expected, p.results, "=>", p.exprStack)

    test("9", 9)
    test("-9", -9)
    test("--9", 9)
    test("-E", -math.e)
    test("9 + 3 + 6", 9 + 3 + 6)
    test("9 + 3 / 11", 9 + 3.0 / 11)
    test("(9 + 3)", (9 + 3))
    test("(9+3) / 11", (9 + 3.0) / 11)
    test("9 - 12 - 6", 9 - 12 - 6)
    test("9 - (12 - 6)", 9 - (12 - 6))
    test("2*3.14159", 2 * 3.14159)




from dataclasses import dataclass, field, fields
from typing import Any

from pprint import pprint
try:
    from icecream import ic
except ImportError:  # Graceful fallback if IceCream isn't installed.
    ic = lambda *a: None if not a else (a[0] if len(a) == 1 else a)  # noqa

from .saxes import SAxes, SAxesModel
from .sgraph import SGraphModel , FillBetween, SGraph
from .lines import VLineModel, VSpanModel, VLine, VSpan
from .types import Size
from .sfigure import SFigure
from .functions import SFunction
from .utils.math_expression import MathExpression
from .converter import AxesAddable, impl_for_type

def to_dict(datainstance:Any) -> dict:
    dd = {f.name: getattr(datainstance, f.name) for f in fields(datainstance)}
    return dd


@dataclass
class AxesWithGraphs:
    axes: SAxesModel
    graphs: list[SGraphModel] = field(default_factory=list)
    areas: list[SGraphModel] = field(default_factory=list)
    vlines: list[VLineModel] = field(default_factory=list)
    vspans: list[VSpanModel] = field(default_factory=list)
    functions: dict[str, MathExpression] = field(default_factory=dict)
    items: list[AxesAddable] = field(default_factory=list)
    left: Size = Size(0)
    bottom: Size = Size(0)
    templates: dict = field(default_factory=dict)
    
    def setup(self, saxes: SAxes):
        for name, func in self.functions.items():
            SFunction(saxes, function=func, name=name)
        for a in self.items:
            impl_for_type(a.class_id)(saxes, **to_dict(a))
        for g in self.graphs:
            SGraph(saxes, **to_dict(g))
            #g.set_saxes(self.axes)
        for a in self.areas:
            FillBetween(saxes, **to_dict(a))
            #a.set_saxes(self.axes)
        for a in self.vlines:
            #pprint(to_dict(a))
            VLine(saxes, **to_dict(a))
            #a.set_saxes(self.axes)
        for a in self.vspans:
            VSpan(saxes, **to_dict(a))
            #a.set_saxes(self.axes)

@dataclass
class FigureDescription:
    figure: SFigure
    axes_descriptors: list[AxesWithGraphs]
    
    def create_figure(self):
        with self.figure:
            #self.figure.set_figure()
            for ax in self.axes_descriptors:
                saxes = SAxes(self.figure, ax.left, ax.bottom, 
                              **to_dict(ax.axes))
                #self.figure.add_axes(ax.axes, left=ax.left, bottom=ax.bottom)
                ax.setup(saxes)    



from dataclasses import dataclass, field
from typing import Optional, Any, ClassVar
import numpy as np

from .converter import register_model, register_impl
from .utils.math_expression import MathExpression

try:
    from icecream import ic
except ImportError:  # Graceful fallback if IceCream isn't installed.
    ic = lambda *a: None if not a else (a[0] if len(a) == 1 else a)  # noqa

from .saxes import SAxes
    
@register_model  
@dataclass
class VLineModel:
    class_id: ClassVar[str] = "VLine"    
    x: MathExpression = MathExpression(0)
    y_min: float = 0.
    y_max: float = 1.
    plot_args: dict[str, Any] = field(default_factory=lambda: dict(alpha=0.5))
@register_impl
class VLine(VLineModel):
    _saxes: SAxes
    def __init__(self, saxes: SAxes, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.set_saxes(saxes)
    def set_saxes(self, saxes: SAxes):
        self._saxes = saxes
        x = self._saxes.evaluate(self.x)
        self._saxes.axes.axvline(x=x,ymin=self.y_min, ymax=self.y_max, 
                                 **self.plot_args)
@register_model  
@dataclass
class HLineModel:
    class_id: ClassVar[str] = "HLine"    
    y: MathExpression = MathExpression(0)
    x_min: float = 0.
    x_max: float = 1.
    plot_args: dict[str, Any] = field(default_factory=lambda: dict(alpha=0.5))
@register_impl
class HLine(HLineModel):
    _saxes: SAxes
    def __init__(self, saxes: SAxes, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.set_saxes(saxes)
    def set_saxes(self, saxes: SAxes):
        self._saxes = saxes
        y = self._saxes.evaluate(self.y)
        self._saxes.axes.axhline(y=y,xmin=self.x_min, xmax=self.x_max, 
                                 **self.plot_args)
@register_model  
@dataclass
class VSpanModel:
    class_id: ClassVar[str] = "VSpan"    
    x0: MathExpression = MathExpression(0)
    x1: MathExpression = MathExpression(0)
    y_min: float = 0
    y_max: float = 1
    plot_args: dict[str, Any] = field(default_factory=lambda: dict(alpha=0.3, lw=0))

@register_impl
class VSpan(VSpanModel):
    _saxes: Optional[SAxes]
    def __init__(self, saxes: SAxes, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.set_saxes(saxes)
    def set_saxes(self, saxes: SAxes):
        self._saxes = saxes
        x0 = self._saxes.evaluate(self.x0)
        x1 = self._saxes.evaluate(self.x1)
        self._saxes.axes.axvspan(xmin=x0, xmax=x1,ymin=self.y_min, 
                                 ymax=self.y_max, **self.plot_args)
        
@register_model  
@dataclass
class HSpanModel:
    class_id: ClassVar[str] = "HSpan"    
    y0: MathExpression = MathExpression(0)
    y1: MathExpression = MathExpression(0)
    x_min: float = 0
    x_max: float = 1
    plot_args: dict[str, Any] = field(default_factory=lambda: dict(alpha=0.3, lw=0))

@register_impl
class HSpan(HSpanModel):
    _saxes: Optional[SAxes]
    def __init__(self, saxes: SAxes, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.set_saxes(saxes)
    def set_saxes(self, saxes: SAxes):
        self._saxes = saxes
        y0 = self._saxes.evaluate(self.y0)
        y1 = self._saxes.evaluate(self.y1)
        self._saxes.axes.axhspan(ymin=y0, ymax=y1,xmin=self.x_min, 
                                 xmax=self.x_max, **self.plot_args)
        
@register_model  
@dataclass
class ArrowModel:
    class_id: ClassVar[str] = "Arrow" 
    x: MathExpression = MathExpression(0)
    y: MathExpression = MathExpression(0)
    dx: MathExpression = MathExpression(0)
    dy: MathExpression = MathExpression(-1)
    width: float = 0.08
    length_includes_head: bool = True
    
    plot_args: dict[str, Any] = field(default_factory=lambda: dict(linewidth=0))

@register_impl
@dataclass
class Arrow(ArrowModel):
    def __init__(self, saxes, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._saxes = saxes
        self.setup_saxes(self._saxes)
    def setup_saxes(self, saxes: SAxes):
        self._saxes = saxes
        x = self._saxes.evaluate(self.x)
        y = self._saxes.evaluate(self.y)
        dx = self._saxes.evaluate(self.dx)
        dy = self._saxes.evaluate(self.dy)
        self._saxes.axes.arrow(x+dx,y+dy, -dx, -dy, 
                               length_includes_head=self.length_includes_head,
                               width=self.width, **self.plot_args)
        


from typing import Union, Protocol, ClassVar, Type, TYPE_CHECKING, TypeVar
try:
    from  typing import dataclass_transform
except ImportError:
    dataclass_transform = lambda: lambda x: x
try:
    from icecream import ic
except ImportError:  # Graceful fallback if IceCream isn't installed.
    ic = lambda *a: None if not a else (a[0] if len(a) == 1 else a)  # noqa

from .utils.math_expression import MathExpression

from .types import FigAction, Size
if TYPE_CHECKING:
    from .saxes import SAxes


from cattrs.preconf.pyyaml import make_converter
def structure_list_of_strings(value, _):
    if isinstance(value, str):
        value = [value]
    return value
def structure_list_of_math_expression(value, _):
    if isinstance(value, str):
        value = [value]
    value = [MathExpression(s) for s in value]
    return value

_converter2 = make_converter(omit_if_default=True)
_converter2.register_structure_hook(Size, lambda val, _: Size(val))
_converter2.register_unstructure_hook(Size, str)
_converter2.register_structure_hook(Union[str, list[str]], structure_list_of_strings)
_converter2.register_structure_hook(MathExpression, lambda val, _: MathExpression(val))
_converter2.register_unstructure_hook(MathExpression, str)
_converter2.register_structure_hook(Union[MathExpression, list[MathExpression]], structure_list_of_math_expression)
_converter2.register_structure_hook(FigAction, lambda val, _: FigAction._member_map_[val])
_converter2.register_unstructure_hook(FigAction, lambda s: s.name)


converter = make_converter(omit_if_default=True)
converter.register_structure_hook(Size, lambda val, _: Size(val))
converter.register_unstructure_hook(Size, str)
converter.register_structure_hook(Union[str, list[str]], structure_list_of_strings)
converter.register_structure_hook(MathExpression, lambda val, _: MathExpression(val))
converter.register_unstructure_hook(MathExpression, str)
converter.register_structure_hook(Union[MathExpression, list[MathExpression]], structure_list_of_math_expression)
converter.register_structure_hook(FigAction, lambda val, _: FigAction._member_map_[val])
converter.register_unstructure_hook(FigAction, lambda s: s.name)



#converter = _converter2.copy()
class AxesAddable(Protocol):
    class_id: ClassVar[str]
class AxesAddableImpl(Protocol):
    class_id: ClassVar[str]
    def __init__(self, saxes: "SAxes", *args, **kwargs): ...    


_modell_class_dict: dict[str, Type[AxesAddable]] = {}
_impl_class_dict: dict[str, Type[AxesAddableImpl]] = {}
def impl_for_type(type: str) -> Type[AxesAddableImpl]:
    return _impl_class_dict[type]

T = TypeVar("T", bound=AxesAddable)
TImpl = TypeVar("TImpl", bound=AxesAddableImpl)

@dataclass_transform() # type: ignore
def register_model(cls: Type[T]) -> Type[T]:
    _modell_class_dict[cls.class_id] = cls
    return cls
@dataclass_transform() # type: ignore
def register_impl(cls: Type[TImpl]) -> Type[TImpl]:
    _impl_class_dict[cls.class_id] = cls
    return cls
def unstructure_AxesAddable(p: AxesAddable, *args):
    dd = _converter2.unstructure(p, *args)
    dd.update({"type": p.class_id})
    return dd
def structure_AxesAddable(data: dict, *args):
    cls = _modell_class_dict[data["type"]]
    return _converter2.structure(data, cls)
converter.register_unstructure_hook(AxesAddable, unstructure_AxesAddable)
converter.register_structure_hook(AxesAddable, structure_AxesAddable)


from dataclasses import dataclass, field
from typing import Optional, Any, TYPE_CHECKING, Callable

from matplotlib.figure import Figure
from matplotlib.axes import Axes
from matplotlib.ticker import MultipleLocator
import numpy as np

from .types import cm, Size, Point

if TYPE_CHECKING:
    from .sfigure import SFigure
    from .utils.math_expression import MathExpression
    from .utils.math_parser import vardict_t
try:
    from icecream import ic
except ImportError:  # Graceful fallback if IceCream isn't installed.
    ic = lambda *a: None if not a else (a[0] if len(a) == 1 else a)  # noqa

        
@dataclass
class SAxesModel:
    x_min: float
    y_min: float
    width: Size = Size(10 * cm)
    height: Size = Size(8 * cm)
    x_label: str = "x"
    y_label: str = "y"
    x_tick_distance: float = 1
    y_tick_distance: float = 1
    unit:Size = Size(1*cm)
    show_legend: bool = True
    legend_options: dict[str, Any] = field(default_factory=dict)    
    n_points = 3000
    x_label_offset: Point = Point(Size(0.5*cm), Size(0))
    y_label_offset: Point = Point(Size(0),Size(0.5*cm))

    def __post_init__(self):
        self.x_max: float = self.x_min + self.width/self.unit
        self.y_max: float = self.y_min + self.height/self.unit
        self._x = np.linspace(self.x_min, self.x_max, self.n_points)
    
class SAxes(SAxesModel):
    axes: Axes
    axes_variables: dict[str, float] 
    axes_functions: dict[str, Callable]
    _sfigure: "SFigure"
    
    def __init__(self, figure: "SFigure", left: Size, bottom: Size, 
                 *args, **kwargs):
        self._sfigure = figure
        #print("XXX", args, kwargs)
        super().__init__(*args, **kwargs)
        self.axes_variables = dict(pi=np.pi, e=np.e)
        self.axes_functions = {}
        self.setup(left, bottom)
        self._sfigure.add_axes(self)
    def get_offset_x_ax(self, dx, dy):
        ax = self.axes
        tr1 = ax.get_yaxis_transform()
        assert ax.figure is not None
        tr2 = ax.figure.dpi_scale_trans.inverted()
        tr = tr1+tr2
        tx,ty = tr.transform((0,0))
        return tr.inverted().transform((tx+dx, ty+dy))
    
    def get_offset_y_ax(self, dx, dy):
        ax = self.axes
        tr1 = ax.get_xaxis_transform()
        assert ax.figure is not None
        tr2 = ax.figure.dpi_scale_trans.inverted()
        tr = tr1+tr2
        tx,ty = tr.transform((0,0))
        return tr.inverted().transform((tx+dx, ty+dy))
    
    def setup(self, left: Size, bottom: Size):
        figure = self._sfigure.figure
        trf = figure.dpi_scale_trans + figure.transFigure.inverted()
        aw, ah = trf.transform((self.width, self.height))
        al, ab = trf.transform((left, bottom))
        rect = (al, ab, aw, ah)
        #ic(rect)
        #ic(self)
        self.axes = figure.add_axes(rect)
        self.axes.set_xlim(self.x_min, self.x_max)
        self.axes.set_ylim(self.y_min, self.y_max)
        
        self.axes.set_aspect("equal")
        self.axes.set_xlabel(self.x_label)
        self.axes.set_ylabel(self.y_label)
        maj_pos_y = MultipleLocator(base=self.x_tick_distance)
        maj_pos_x = MultipleLocator(base=self.y_tick_distance)

        self.axes.xaxis.set(major_locator=maj_pos_x)
        self.axes.yaxis.set(major_locator=maj_pos_y)
        self.axes.tick_params(axis='both', which='minor', length=0)   # remove minor tick lines

        spine_x = self.axes.spines["bottom"]
        spine_y = self.axes.spines["left"]
        spine_x.set_position(('data', 0))
        spine_y.set_position(('data', 0))
        self.axes.plot(1, 0, ">k", transform=self.axes.get_yaxis_transform(), clip_on=False)
        self.axes.plot(0, 1, "^k", transform=self.axes.get_xaxis_transform(), clip_on=False)

        self.axes.spines['top'].set_visible(False)
        self.axes.spines['right'].set_visible(False)

        dx, dy = self.x_label_offset.args()
        dx1, dy1 = self.get_offset_x_ax(dx, dy)
        self.axes.xaxis.set_label_coords(1+dx1, 0+dy1, 
                                         transform=self.axes.get_yaxis_transform())
        self.axes.yaxis.label.set(rotation="horizontal")
        dx, dy = self.y_label_offset.args()
        dx1, dy1 = self.get_offset_y_ax(dx, dy)
        self.axes.yaxis.set_label_coords(0+dx1, 1+dy1, 
                                         transform=self.axes.get_xaxis_transform())
        
    def evaluate(self, 
                 expr: "MathExpression", 
                 vars: Optional["vardict_t"] = None, 
                 functions: Optional["vardict_t"]= None):
        if vars is None:
            vars = {}
        vars = vars | self.axes_variables
        if functions is None:
            functions = {}
        functions = functions | self.axes_functions
        try:
            res = expr.evaluate(vars, functions)
        except Exception as e:
            print(f"Warning: got exception '{e}' in evaluation of '{expr}'.")
            res = 0
        return res
        
    
    def finalize(self):
        if self.show_legend:
            legend=self.axes.legend(**self.legend_options)
            legend.set_zorder(30)

    def plot(self, *args, **kwargs):
        return self.axes.plot(*args, **kwargs)
    def legend(self, *args, **kwargs):
        return self.axes.legend(*args, **kwargs)
        




from .converter import converter
from .figure_description import FigureDescription
import argparse
from os import path
from typing import Optional

try:
    from icecream import ic
except ImportError:  # Graceful fallback if IceCream isn't installed.
    ic = lambda *a: None if not a else (a[0] if len(a) == 1 else a)  # noqa

def generate_plot(infile: str, outfile: Optional[str] = None):
    with open(infile, "r") as f:
        fdsc = converter.loads(f.read(), FigureDescription)
        fdsc.figure.output_file = outfile
        fdsc.create_figure()
def main():
    parser = argparse.ArgumentParser(
        prog="schulplots.py", 
        description="Create function plots styled similar to the conventions used in German schools.\n"\
            "For documentation, see https://schulplots.hans-aschauer.de"
    )
    parser.add_argument("filename", help="Description of the figure, in YAML format")
    parser.add_argument("--output", "-o", 
                        help="Name out output file. If not provided (and --show is "\
                             "not given), store the figure in the current working "\
                             "directory.")
    parser.add_argument("--show", "-s", help="Show plot in interactive window. May not be used with --output.",
                        action="store_true", default=False)

    args = parser.parse_args()
    if args.output is not None and args.show:
        parser.print_help()
    elif args.output is not None and not args.show:
        generate_plot(args.filename, args.output)
        #fdsc.figure.output_file = args.output
    elif args.output is None and args.show:
        generate_plot(args.filename, None)
        #fdsc.figure.output_file = None
    elif args.output is None and not args.show:
        bname, _ = path.splitext(path.basename(args.filename))
        generate_plot(args.filename, bname + ".png")
        #fdsc.figure.output_file = bname + ".png"
        
    else:
        raise ValueError("Logic error: this should not happen")
    #fdsc.create_figure()
        



from dataclasses import dataclass
from typing import Union
import re

try:
    from icecream import ic
except ImportError:  # Graceful fallback if IceCream isn't installed.
    ic = lambda *a: None if not a else (a[0] if len(a) == 1 else a)  # noqa

from enum import Enum, auto
class FigAction(Enum):
    INTERACT = auto()
    SAVE = auto()

cm = 1./2.54 

size_re= re.compile(r"(-?[\d\.]+)(\s*)([a-z]+)")
units = {"cm": 1./2.54, 
         "mm": 0.1/2.54,
         "in": 1.}
#%%

class Size(float):
    def __new__(cls, value: Union[float, str]):
        if isinstance(value, str):
            if value.isnumeric():
                value = float(value)
            else:
                m = size_re.match(value)
                if m is None: 
                    raise ValueError("Invalid size")
                val, _, unit = m.groups()
                value = float(val) * units[unit]
        return float.__new__(cls, value)
    def __init__(self, value: Union[float, str]):
        self._init_as = str(value)
    def __repr__(self):
        return self._init_as
    __str__ = __repr__

@dataclass(frozen=True)
class Point:
    x: Size
    y: Size
    def args(self):
        return self.x, self.y
